%top{
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ruby.h>
}

%{
static VALUE token(const char* name, const char* txt, size_t len)
{
        VALUE tok = rb_ary_new2(2);
        rb_ary_store(tok, 0, ID2SYM(rb_intern(name)));
        rb_ary_store(tok, 1, rb_str_new(txt, len));
        return tok;
}

static VALUE token_text(const char* name)
{
        return token(name, yytext, yyleng);
}

static VALUE token_empty(const char* name)
{
        return token(name, NULL, 0);
}

static VALUE end_lex()
{
        VALUE end_arr = rb_ary_new2(2);
        rb_ary_store(end_arr, 0, Qfalse);
        rb_ary_store(end_arr, 1, rb_str_new2("$end"));
        return end_arr;
}

#define YY_DECL VALUE yylex()

#define yyterminate() return end_lex()
%}

%option noyywrap

%%

Vk[a-zA-Z0-9]+    return token_text("STRC");
vkCmd[a-zA-Z0-9]+ return token_text("CMD");
vk[a-zA-Z0-9]+    return token_text("FN");
[a-zA-Z0-9_]+     return token_text("NONVK");
\(.+\)            return token_text("ANNOT");

^[ ]{4}        return token_empty("CNCT");
\[r\]          return token_empty("RET");
\[o\]          return token_empty("OPT");
\[a\]          return token_empty("ARR");
\[m\]          return token_empty("MSK");

[ \t]+
\n

.              rb_raise(rb_eRuntimeError, "unrecognized: %s", yytext);

%%

static void handle_cerr(const char* msg)
{
        char* errdesc = strerror(errno);
        rb_raise(rb_eRuntimeError, "%s: %s", msg, errdesc);
}

static VALUE lexer_init(VALUE self, VALUE fname_rb)
{
        if (TYPE(fname_rb) != T_STRING) {
                rb_raise(rb_eTypeError, "filename must be a string");
        }

        const char* fname = StringValueCStr(fname_rb);

        yyin = fopen(fname, "r");

        if (yyin == NULL) {
                handle_cerr("fopen");
        }

        return self;
}

static VALUE lexer_lex(VALUE self)
{
        /*
        char*  buf;
        size_t len;

        yyout = open_memstream(&buf, &len);
        if (yyout == NULL) {
                handle_cerr("open_memstream");
        }
        */

        /*yylex();*/

        /*
        if (fflush(yyout) == EOF) {
                handle_cerr("fflush");
        }

        VALUE str = rb_str_new(buf, len);

        free(buf);
        */

        return yylex();
}

void Init_lexer(void)
{
        VALUE cVkGraph = rb_const_get(rb_cObject, rb_intern("VkGraph"));
        VALUE cLexer   = rb_define_class_under(cVkGraph, "Lexer", rb_cObject);

        rb_define_method(cLexer, "initialize", lexer_init, 1);
        rb_define_method(cLexer, "lex",        lexer_lex,  0);
}
